

	/*********************************************************************
	 * These routines set up a series of simultaneous equations which 
	 * correspond to the photometric solution of a large number
	 * of observations.  We set up a big matrix equation of the form
	 *
	 *              A  *  x  =   b
	 *
	 * where 'A' is a matrix, 'b' is a vector with the observed values of
	 * instr. mags and weights, and 'x' is the unknown vector of
	 * magnitude adjustments that minimizes the sums of the squares
	 * of the residuals from the mean. 
	 *
	 * Then, we call some Numerical Recipes code is used to solve it.
	 *
	 * MWR 10/18/2003 
	 *
	 *
	 * Added FORCE_STAR_MAG option, which may help solution under
	 *   some circumstances.
	 * Added a shift to the image adjustment values so that the
	 *   input (raw) mags and output (true) mags do obey the relationship
	 *          true_mag   +   image_adj   =   raw_mag
	 *
	 * MWR 10/28/2003
	 *
	 * Modified code to include terms for zero-point offsets of each
	 * filterchip.  That means that we now solve for magnitude of each
	 * star, adjustment of each image, and offset of each chip.
	 *
	 * MWR 3/23/2004
	 *
	 * Fixed huge error in "zero_vector()" which actually prevented
	 * any elements from being set to zero (!).
	 * MWR 3/25/2004
	 *
	 * Commented out the "calc_model_mag()" function, since it  
	 * isn't used any more in this simplified version.
	 * MWR 11/8/2004
	 *
	 * Replaced all the Numerical Recipes routines with equivalents
	 * from the GNU Scientific Library.  This means vector and matrix
	 * allocation and de-allocation, plus functions to do the SVD
	 * analysis of the matrix and then apply the result to find the
	 * solution vector.
	 * The big pain was changing all the old 1-index code to adhere
	 * to the GSL 0-index convention for vectors and matrices.
	 * MWR 12/13/2004
	 *
	 *
	 * Fixed bug in "find_sigm()".  Code which used neighboring stars
	 *     to estimate uncertainty did not check to make sure that
	 *     the neighboring star's uncertainty was valid, so some values
	 *     of NODATA = 99.0 were being included in calculations.
	 * MWR 4/8/2006
	 *
	 * More bugs in "find_sigm()".  Not really a bug: in sparse fields,
	 *     there may be no other stars in the field within MAG_MATCH 
	 *     mag of the variable star -- which means no estimation of 
	 *     uncertainty based on similar stars is possible.  Added some
	 *     DEBUG statements to find the problem.  The solution is simply
	 *     to increase the MAG_MATCH value from default of 0.30 mag
	 *     to some larger value in sparse fields.  
	 * MWR 4/12/2006
	 *
	 * Added "mag_match" arg to find_sigm(), which can be set via
	 *     optional command-line arguement.  Replaces the 
	 *     fixed MAG_MATCH value.
	 * MWR 5/9/2006
	 *   
	 * Major changes to allow for a new execution mode:
	 *     instead of using all stars in all images to determine the 
	 *     solution, use only a small number of bright ones; then 
	 *     apply that solution to all stars.
	 *
	 *     - new argument 'n_quick' to solvematrix
	 *           if == 0, use all stars, as in the past
	 *           if > 0, use only the n_quick brightest stars
	 *
	 *     - now pass as arguments the "image_arr[]" and "star_arr[]"
	 *           arrays, so that we can either pass the full
	 *           arrays with all stars, or a smaller subset 
	 *           which will run more quickly
	 *
	 *     - in 'quick' mode, we create copies of the original 
	 *           "image_arr[]" and "star_arr[]" arrays;
	 *           the new "star_arr[]" copy includes only the 
	 *           brightest 'n_quick' stars in it.  We use the
	 *           copies for the main matrix calculations, then
	 *           transfer the results back into the original,
	 *           full-sized copies for the final stages of analysis
	 *           and output.
	 *            
	 * MWR 6/7/2019
	 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "solvepht.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_linalg.h"

#define TOLERANCE    1.0e-7
#define TINY         (2.0*TOLERANCE) 
#define MAG_MATCH    0.60		/* used by find_sigm() */

#undef DEBUG

	/* 
	 * We use these structures in the find_sigm() and associated
	 *    functions.  They help us to locate stars within
	 *    a narrow range of magnitudes.
	 */
struct s_sort_mag {
	int star_id;
	double mag0;
	double sigm0;
};

  /* 
	* if this is defined, then we set matrix elements corresponding 
	*   to the first star in the list to be zero.   That means that
	*   the first star's mag in the output should be zero, too.
	*   After solving the matrix equation, we need to go back and 
	*   set the output vector x element for this first star to be
	*   zero.  And _then_ we can run the normalize() function,
	*   which will shift the output magnitude values so that the
	*   BRIGHTEST star has average mag zero, rather than the first
	*   star in the list.
	*
	* Setting this option may cause the matrix solution to behave
	*   better.  In some cases, however, the code doesn't need us
	*   to take this step; it returns a reasonable solution on its
	*   own.
	*/
#define FORCE_STAR_MAG 



static void init_matrix(int full_num_images, int full_num_stars, int ee, int ss);
static void calc_em(struct s_image *image_arr, int m, int n);
static void calc_mags(struct s_image *image_arr, int m, int n);
static void calc_sigmas(struct s_image *image_arr, 
             struct s_star *star_arr, int m, int n, double mag_match);
static void zero_vector(gsl_vector *p, int m, double tolerance);
static void print_vector(gsl_vector *p, int m);
static void print_matrix(gsl_matrix *p, int m, int n);
static void fill_bvec(gsl_vector *b, int m, int n);
static void read_data(struct s_star *star_arr, int ee, int ss);
static void fill_matrix(gsl_matrix *p, int ee, int ss);
static void copy_matrix(gsl_matrix *p, gsl_matrix *q, int m, int n);
static void normalize(int m, int n, double value);
static void find_means(struct s_star *star_arr, int m, int n);
static double find_sigm(int n, int e, int s, double mag_match, 
                        int star_id, 
                        struct s_sort_mag *smag_array, int num_sort);
static struct s_sort_mag * build_sort_mag_array(struct s_star *star_arr,
                                                 int num_stars,
                                                 int *num_in_array);
static int find_neighbor_indices(struct s_sort_mag *smag_array,
                                 int num_stars, double this_mag,
                                 double mag_range, 
                                 int *start_index, int *end_index);
static void print_curves(int m, int n);
static void print_sigmas(struct s_image *image_arr,
                         struct s_star *star_arr, int m, int n);
static double calc_model_correction(int num_images, int num_stars, 
		                        int image_index, int star_index);
#if 0
static double calc_model_mag(int num_images, int num_stars, 
		                        int image_index, int star_index);
#endif

static struct s_image *copy_image_array (struct s_image *orig_array, 
                                  int num_orig_images);
#ifdef DEBUG
static void print_image_array(struct s_image *im_array,
                                  int num_images);
#endif
static struct s_star *copy_star_subset(struct s_star *orig_array,  
                                  int num_orig_stars, int n_bright);
static int comp_double(double *d1, double *d2);
static int comp_sortmag(struct s_sort_mag *s1, struct s_sort_mag *s2);
static double calc_cutoff_mag(struct s_star *star_array,
                                 int num_stars, int n_bright);
static int copy_one_star(struct s_star *from, struct s_star *to);
static int weed_image_array(struct s_image *image_array, int num_images,
                             struct s_star *star_array, int num_stars,
                             int num_original_stars);
static int transfer_solution(struct s_image *full_image_arr,
                             int full_num_images,
                             struct s_star *full_star_array,
                             int full_num_stars,
                             struct s_image *q_image_array,
                             struct s_star *q_star_array, 
                             int q_num_stars);


	/* these are inputs */
static gsl_matrix *mag;          /* ee by ss matrix of instr. mags */
static gsl_matrix *weight;			/* ee by ss matrix of weights */

	/* these are computational values */
static gsl_matrix *a;				/* ee+ss+cc by ee+ss+cc matrix; A above */
static gsl_matrix *u;				/* ee+ss+cc by ee+ss+cc: part 1 of the SVD output */
static gsl_vector *w;				/* ee+ss+cc vector: part 2 of SVD output */
static gsl_matrix *v;				/* ee+ss+cc by ee+ss+cc: part 3 of SVD output */
static gsl_vector *b;				/* ee+ss+cc vector: 'b' above */
static gsl_vector *x;				/* ee+ss+cc vector: 'x' above, the 'solution' */
	/* and these are output results */
static gsl_vector *em;				/* ee vector: mean correction for image e */
static gsl_vector *mag0;				/* ss vector: mean mag of star s */
static gsl_vector *sigm0;			/* ss vector: sigma of mean value of star s */
static gsl_matrix *Mes;				/* ee by ss matrix: solved mag of star s */
                   				/*    on image e */
static gsl_matrix *sigMes;			/* ee by ss matrix: total sigma of */
                  				/*    solution's value of star s on image e */
static gsl_vector *sigem;			/* ee vector: sigma of exp corr. of image e */
static gsl_vector *sigemn;			/* ee vector: sigem/# stars in image e */



	/*********************************************************************
	 * PROCEDURE: solve_matrix
	 *
	 * DESCRIPTION: set up all the equations and solve them.
	 *              If the debug flag is set, print out some results.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
solve_matrix
	(
	double zero_point,      /* I: set brightest star to this mag on output */
	double mag_match,       /* I: used to find stars "similar" enough to */
	                        /*       a variable to estimate uncert in its */
	                        /*       individual mag measurements */
	int n_quick             /* I: if > 0, we create a subset of all stars */
	                        /*       using just n_quick brightest ones,   */
	                        /*       solve quickly, then apply solution   */
	                        /*       to all stars                         */
	)
{
	int i;
	int ee, ss;
	int n_bright;
	int active_num_images, active_num_stars;
	double tolerance;
	gsl_vector *work;
	struct s_image *q_image_array, *active_image_arr;
	struct s_star *q_star_array, *active_star_arr;

#ifdef DEBUG
	printf("entering solve_matrix: \n");
	printf("   zero_point    %6.3f \n", zero_point);
	printf("   mag_match     %6.3f \n", mag_match);
	printf("   n_quick       %6d \n", n_quick);
#endif

	/* 
 	 * Check to see if there are more than n_quick stars in the input.
 	 *     If so, we will only use the brightest n_quick stars
 	 *              to solve for image-to-image variations.
 	 *                  n_bright = n_quick
 	 *
 	 *     If not, then use all the stars; we signal this by
 	 *              setting n_bright to zero
 	 *                  n_bright = 0
	 */
	if (n_quick < num_stars) {
		/* create a bright subset to calculation solution */
		n_bright = n_quick;
	}
	else {
		/* we will use all stars */
		n_bright = 0;
	}

	/*
 	 * If the user requests that we use a subset of stars for 
 	 *   the solution, make copies of the full arrays of
 	 *   all images and all stars.  We'll use the copies
 	 *   for the quick solution, then use those quick results
 	 *   to fill values in the full arrays later.
 	 */
	if (n_bright > 0) {
		q_image_array = copy_image_array(image_arr, num_images);
		if (q_image_array == NULL) {
			fprintf(stderr, "copy_image_array fails \n");
			exit(1);
		}
#ifdef DEBUG
		/* print the original array, then the copy, so we can compare */
		print_image_array(image_arr, num_images);
		printf("Q\n");
		print_image_array(q_image_array, num_images);
#endif

		/* create a subset of the n_bright brightest stars */
		q_star_array = copy_star_subset(star_arr, num_stars, n_bright);

		/* weed out all non-bright star measures from new_image_array */
		if (weed_image_array(q_image_array, num_images,
							q_star_array, n_bright, num_stars) != 0) {
			fprintf(stderr, "weed_image_array fails \n");
			exit(1);
		}

		/* 
 		 * the "active" vars are the ones we'll pass to the routines
 		 *    up to and including the linear algebra.  Since we are
 		 *    using 'quick' mode, set these to the bright-only subset
 		 */
		active_image_arr = q_image_array;
		active_star_arr = q_star_array;
		active_num_images = num_images;
		active_num_stars = n_bright;
	}
	else {
		/* 
 		 * we will use all stars in the solution, so set "active"
		 *    vars to the full-size versions.
		 */
		active_image_arr = image_arr;
		active_star_arr = star_arr;
		active_num_images = num_images;
		active_num_stars = num_stars;
	}
	

#define ACTIVE

	tolerance = TOLERANCE;
#ifdef ACTIVE
	ee = active_num_images;
	ss = active_num_stars;
#else
	ee = num_images;
	ss = num_stars;
#endif
	/* 
 	 *  we must allocate space for full-size vectors, even if we
 	 *     are going to use a subset for some calculations
 	 */
	init_matrix(num_images, num_stars, ee, ss);


#ifdef ACTIVE
	read_data(active_star_arr, ee, ss);
#else
	read_data(star_arr, ee, ss);
#endif
	fill_matrix(a, ee, ss);
	if (debug_flag) {
		printf("\n This is the matrix \n");
		print_matrix(a, ee + ss, ee + ss);
	}
	copy_matrix(a, u, ee + ss, ee + ss);

	fill_bvec(b, ee, ss);
	if (debug_flag) {
		printf(" This is the B vector \n");
		print_vector(b, ee + ss);
	}

#if 0
	svdcmp(u, ee + ss, ee + ss, w, v);
#endif
	if ((work = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "solve_matrix: can't allocate for vector work\n");
		exit(-1);
	}
	if (gsl_linalg_SV_decomp(u, v, w, work) != 0) {
		fprintf(stderr, "solve_matrix: gsl_linalg_SV_decomp returns with error\n");
		exit(1);
	}

	if (debug_flag) {
		printf("this is the svd matrix U \n");
		print_matrix(u, ee + ss, ee + ss);
		printf("this is the vector of w(j) - which could need zero'ing \n");
		print_vector(w, ee + ss);
	}

	zero_vector(w, ee + ss, tolerance);
	if (debug_flag) {
		printf("this is the vector w(j) after zeroing tiny elements\n");
		print_vector(w, ee + ss);
	}

#if 0
	svbksb(u, w, v, ee + ss, ee + ss, b, x);
#endif
   if (gsl_linalg_SV_solve(u, v, w, b, x) != 0) {
		fprintf(stderr, "solve_matrix: gsl_linalg_SV_solve returns with error\n");
		exit(1);
	}

	if (debug_flag) {
		printf("\nand this is the vector of x - the solution\n");
		print_vector(x, ee + ss);
	}

#ifdef FORCE_STAR_MAG
	/* 
	 * the first star's mag in the solution vector x has been unconstrained
	 *   if we chose the FORCE_STAR_MAG option.  We need to set it equal
	 *   to zero explicitly before we go any further.
	 */
	gsl_vector_set(x, ee, 0.0);
#endif

	/*
 	 * now, if necessary, we transfer information from the small
 	 *    "active" arrays back to the full-sized arrays,
 	 *    so that magnitudes can be computed for all stars.
 	 */
	if (n_bright > 0) {
		if (transfer_solution(image_arr, num_images, star_arr, num_stars,
						q_image_array, q_star_array, n_bright) != 0) {
			fprintf(stderr, "transfer_solutions fails \n");
			exit(1);
		}

		/* and go back to using the full-sized arrays in all calculations */
		active_image_arr = image_arr;
		active_star_arr = star_arr;
		active_num_images = num_images;
		active_num_stars = num_stars;

		ee = num_images;
		ss = num_stars;
	}
	

#ifdef ACTIVE
	calc_em(active_image_arr, ee, ss);
	calc_mags(active_image_arr, ee, ss);
	find_means(active_star_arr, ee, ss);
	calc_sigmas(active_image_arr, active_star_arr, ee, ss, mag_match);
#else
	calc_em(image_arr, ee, ss);
	calc_mags(image_arr, ee, ss);
	find_means(star_arr, ee, ss);
	calc_sigmas(image_arr, star_arr, ee, ss, mag_match);
#endif
	normalize(ee, ss, zero_point);

	if (debug_flag) {
		print_curves(ee, ss);
#ifdef ACTIVE
		print_sigmas(active_image_arr, active_star_arr, ee, ss);
#else
		print_sigmas(image_arr, star_arr, ee, ss);
#endif
	}

	/* 
	 * set the "star->mag0" and "star->sigma" fields to the 
	 *   mean magnitude and scatter-around-mean-magnitude
	 *   in the solution.
	 */
#ifdef ACTIVE
	for (i = 0; i < active_num_stars; i++) {
		if (active_star_arr[i].include != 0)  {
			active_star_arr[i].mag0 = gsl_vector_get(mag0, i);
			active_star_arr[i].sigma = gsl_vector_get(sigm0, i);
		}
	}
#else
	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include != 0)  {
			star_arr[i].mag0 = gsl_vector_get(mag0, i);
			star_arr[i].sigma = gsl_vector_get(sigm0, i);
		}
	}
#endif


}


	/************************************************************************
	 * PROCEDURE: init_matrix
	 *
	 * DESCRIPTION: allocate space for all the 1-indexed arrays and matrices
	 *              that will be needed to set up and solve the series
	 *              of linear equations.
	 *
	 *              The working matrices and vectors used in the 
	 *              least-squares calculations may be either full-sized,
	 *              or smaller if the user wants to perform a quick
	 *              solution; but the vectors used for output quantities
	 *              must always be full-sized.
	 *
	 * RETURNS:
	 *    nothing
	 */


static void
init_matrix
	(
	int full_num_images, /* I: full number of images in dataset */
	int full_num_stars,  /* I: full number of stars in dataset */
	int ee,              /* I: number of images in solution calcs */
	int ss               /* I: number of stars in solution calcs */
	)
{


	/* 
 	 * these quantities are used in least-squares calculations,
 	 *   and may be full-sized, or smaller, depending on user's choice
 	 */
	if ((mag = gsl_matrix_calloc(ee, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix mag\n");
		exit(-1);
	}
	if ((weight = gsl_matrix_calloc(ee, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix weight\n");
		exit(-1);
	}
	if ((a = gsl_matrix_calloc(ee + ss, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix a\n");
		exit(-1);
	}
	if ((b = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector b \n");
		exit(-1);
	}
	if ((u = gsl_matrix_calloc(ee + ss, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix u\n");
		exit(-1);
	}
	if ((v = gsl_matrix_calloc(ee + ss, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix v\n");
		exit(-1);
	}
	if ((w = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector w \n");
		exit(-1);
	}
	if ((x = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector x \n");
		exit(-1);
	}

	/* all of these are used for output quantities, must be full size */
	if ((em = gsl_vector_calloc(full_num_images)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector em \n");
		exit(-1);
	}
	if ((mag0 = gsl_vector_calloc(full_num_stars)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector mag0 \n");
		exit(-1);
	}
	if ((sigm0 = gsl_vector_calloc(full_num_stars)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigm0 \n");
		exit(-1);
	}
	if ((Mes = gsl_matrix_calloc(full_num_images, full_num_stars)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix Mes \n");
		exit(-1);
	}
	if ((sigMes = gsl_matrix_calloc(full_num_images, full_num_stars)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix sigMes \n");
		exit(-1);
	}
	if ((sigem = gsl_vector_calloc(full_num_images)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigem \n");
		exit(-1);
	}
	if ((sigemn = gsl_vector_calloc(full_num_images)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigemn \n");
		exit(-1);
	}
}



	/**********************************************************************
	 * PROCEDURE: calc_sigmas
	 *
	 * DESCRIPTION: After we have solved the equations, we can go back
	 *              and calculate values of the RMS for each exposure 
	 *              (as a whole), for mean magnitude of each star, 
	 *              and for each observation of each star (individually).
	 *
	 *              The tricky part is to handle values of NODATA 
	 *              correctly so that they don't throw averages off. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_sigmas
	(
	struct s_image *image_arr,   /* I: array of images */
	struct s_star *star_arr,     /* I: array of stars */
	int m,              /* I: number of images in solution */
	int n,              /* I: number of stars in solution */
	double mag_match    /* I: pass to find_sigm() function */
	)
{
	int i, j, num, num_sort;
	double temp, temp2, top, bot;
	struct s_measurep *mp;
	struct s_measure *measure;
	struct s_sort_mag *sort_mag_array;


	/* set everything to 0.0, first of all */
	for (i = 0; i < m; i++) {
		gsl_vector_set(sigem, i, 0.0);
		gsl_vector_set(sigemn, i, 0.0);
		for (j = 0; j < n; j++) {
			gsl_matrix_set(sigMes, i, j, 0.0);
		}
	}
	for (i = 0; i < n; i++) {
		gsl_vector_set(sigm0, i, 0.0);
	}

	/* first find sigma(em(e)) */
	if (n >= 2) {
		for (i = 0; i < m; i++) {
			num = 0;
/****
			for (sp = image_arr[i].starp; sp != NULL; sp = sp->next)
				if (sp->star->include != 0)
					num++;
***/
			top = 0.0;
			bot = 0.0;
			for (j = 0; j < n; j++) {
				if ((gsl_matrix_get(Mes, i, j) == NODATA) || 
					 (gsl_vector_get(mag0, j) == NODATA)) {
					continue;
				}
				num++;
				temp = (gsl_matrix_get(Mes, i, j) - gsl_vector_get(mag0, j));
				top += temp*temp*gsl_matrix_get(weight, i, j);
				bot += gsl_matrix_get(weight, i, j);
			}
			if (bot > 0.0) {
				temp = sqrt((n*top)/((n - 1)*bot));
				gsl_vector_set(sigem, i, temp);
				if (num > 0) {
					temp = gsl_vector_get(sigem, i)/(double) sqrt((double) num);
					gsl_vector_set(sigemn, i, temp);
				}
			}
			else {
				gsl_vector_set(sigem, i, NODATA);	
				gsl_vector_set(sigemn, i, NODATA);	
			}
		}
	}

	/* now, find sigma(m0(s)) */
	if (m >= 2) {
		for (i = 0; i < n; i++) {
			top = 0.0;
			bot = 0.0;
			if (star_arr[i].variable == 0) {
				for (j = 0; j < m; j++) {
					if ((gsl_matrix_get(Mes, j, i) == NODATA) || 
					               (gsl_vector_get(mag0, i) == NODATA) ||
										(image_arr[j].include == 0)) {
						continue;
					}
					temp = gsl_matrix_get(Mes, j, i) - gsl_vector_get(mag0, i);
					top += temp*temp*gsl_matrix_get(weight, j, i);
					bot += gsl_matrix_get(weight, j, i);
				}
			}
			else {
				for (mp = star_arr[i].mp; mp != NULL; mp = mp->next) {
					measure = mp->measure;
					shAssert(measure != NULL);
					j = measure->image_id;
					if ((gsl_matrix_get(Mes, j, i) == NODATA) || 
					               (gsl_vector_get(mag0, i) == NODATA) ||
										(image_arr[j].include == 0)) {
						continue;
					}
					temp = gsl_matrix_get(Mes, j, i) - gsl_vector_get(mag0, i);
					top += temp*temp*measure->w4;
					bot += measure->w4;
				}
			}
			if (bot > 0.0) {
				temp = sqrt((m*top)/((m - 1)*bot));
				gsl_vector_set(sigm0, i, temp);
			}
			else {
				gsl_vector_set(sigm0, i, NODATA);
			}
		}
	}

	/*
 	 * we'll use this in the next step of computing sigma(M(e,s))
 	 *     each time we call find_sigm()
 	 */
	if ((sort_mag_array = build_sort_mag_array(star_arr, n, &num_sort)) == NULL) {
		fprintf(stderr, "calc_sigmas: build_sort_mag_array fails \n");
		exit(1);
	}

	/* now we can finally calculate sigma(M(e,s)) */
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			int this_id = star_arr[j].id;
			temp = find_sigm(n, i, j, mag_match, this_id, sort_mag_array, num_sort);
			if (temp == NODATA) {
				gsl_matrix_set(sigMes, i, j, NODATA);
			}
			else if ((temp2 = gsl_vector_get(sigemn, i)) != NODATA) {
				temp = sqrt(temp*temp + temp2*temp2);
				gsl_matrix_set(sigMes, i, j, temp);
			}
			else {
				gsl_matrix_set(sigMes, i, j, temp);
			}
		}
	}

	/* should free the sort mag array here */
	free(sort_mag_array);
}


	/************************************************************************
	 * PROCEDURE: print_sigmas
	 *
	 * DESCRIPTION: print out some derived quantities.  For debugging
	 *              purposes only.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_sigmas
	(
	struct s_image *image_arr,    /* I: array of images */
	struct s_star *star_arr,      /* I: array of stars */
	int m,                        /* I: number of images */
	int n                         /* I: number of stars */
	)
{
	int i;

	/* first find sigma(em(e)) */
	printf("   Sigma of Exposures: image #, sigem, sigemn\n");
	for (i = 0; i < m; i++) {
		printf(" %4d  %5.3f  %5.3f\n", image_arr[i].index, 
							 gsl_vector_get(sigem, i), 
							 gsl_vector_get(sigemn, i)); 
	}
	printf("\n");

	/* now, find sigma(m0(s)) */
	printf("  Sigma of stellar mean mags: star #, m0, sig(m0) \n");
	for (i = 0; i < n; i++) {
		printf("  %6d %6.3f %5.3f  \n", star_arr[i].id, 
				  			gsl_vector_get(mag0, i), 
							gsl_vector_get(sigm0, i)); 
	}
	printf("\n");
}



	/***********************************************************************
	 * PROCEDURE: calc_em
	 *
	 * DESCRIPTION: Calculate the 'em' matrix - the overal shift of 
	 *              each image's magnitudes to bring them close to the mean 
	 *              values. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_em
	(
	struct s_image *image_arr,   /* I: array of images */
	int m,                       /* I: number of images */
	int n                        /* I: number of stars */
	)
{
	int i;
	
	for (i = 0; i < m; i++) { 
		if (image_arr[i].include) {
			gsl_vector_set(em, i, gsl_vector_get(x, i));
		}
		else {
			gsl_vector_set(em, i, NODATA);
		}
	}
}


	/***********************************************************************
	 * PROCEDURE: calc_mags
	 *
	 * DESCRIPTION: Adjust the instrumental magnitudes for each star, 
	 *              adding the results of the matrix inversion to the
	 *              raw instrumental magnitude.  The result is the
	 *              differential magnitude of the star relative to
	 *              all the others in all images.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_mags
	(
	struct s_image *image_arr,   /* I: array of images */
	int m,                       /* I: number of images */
	int n                        /* I: number of stars */
	)
{
	int i, j;	
	double model_correction;
	double temp;
	
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			if ((gsl_matrix_get(mag, i, j) != NODATA) && 
			    (image_arr[i].include != 0)) {
				model_correction = calc_model_correction(m, n, i, j);
#ifdef DEBUG
				printf ("  calc_mags: %5d %5d  %6.3lf %6.3lf   %6.3lf \n",
						i, j, 
						gsl_matrix_get(mag, i, j), 
						model_correction, 
						gsl_matrix_get(mag, i, j) - model_correction);
#endif
				if (model_correction != NODATA) {
					temp = gsl_matrix_get(mag, i, j) - model_correction;
					gsl_matrix_set(Mes, i, j, temp);
				}
				else {
					gsl_matrix_set(Mes, i, j, NODATA);
				}
			}
			else {
				gsl_matrix_set(Mes, i, j, NODATA);
			}
		}
	}
}





	/************************************************************************
	 * PROCEDURE: print_curves
	 *
	 * DESCRIPTION: Print out light curves for each star 
	 *              (used only in debugging)
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_curves
	(
	int m,             /* I: number of images */
	int n              /* I: number of stars */
	)
{
	int i, j;	
	
	printf("\n");
	for (i = 0; i < n; i++) {
		printf("star   %2d: ", i);
		for (j = 0; j < m; j++) {
			printf("%6.3f %6.3f  ", gsl_matrix_get(Mes, j, i), 
								 gsl_matrix_get(sigMes, j, i));
		}
		printf("\n");
	}
}




	/************************************************************************
	 * PROCEDURE: zero_vector
	 *
	 * DESCRIPTION: Set to zero all elements of vector p which are 
	 *              less than 'tolerance' times the maximum element of p
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
zero_vector
	(
	gsl_vector *p,        /* I/O: a vector whose elements we may set to zero */
	int m,                /* I: number of elements in vector */
	double tolerance      /* I: if elem i < this times max elem, set to 0  */
	)
{
	int i;
	double max, min;

	max = 0.0;
	for (i = 0; i < m; i++) {
		if (gsl_vector_get(p, i) > max) {
			max = gsl_vector_get(p, i);
		}
	}
	min = tolerance*max;
	for (i = 0; i < m; i++) {
		if (gsl_vector_get(p, i) < min) {
			gsl_vector_set(p, i, 0.0);
		}
	}
}	



	/*******************************************************************
	 * PROCEDURE: print_vector
	 *
	 * DESCRIPTION: Print a vector (which goes 1-m) in a nice format
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_vector
	(
	gsl_vector *p,      /* I: the vector we'll print */
   int m               /* I: number of elements in the vector */
	)
{
	int i;

	for (i = 0; i < m; i++) {
		printf("%9.3e ", gsl_vector_get(p, i));
	}
	putchar('\n');
}


	/**********************************************************************
	 * PROCEDURE: print_matrix
	 *
	 * DESCRIPTION: Print a matrix (which goes 1-m, 1-n) in a nice format 
	 *
	 * RETURNS:
	 *   nothing
	 */

static void
print_matrix
	(
	gsl_matrix *p,         /* I: matrix whose elements we print */
	int m,                 /* I: number of rows in matrix */
	int n                  /* I: number of columns in vector */
	)
{
	int i, j;

	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			printf("%9.3e ", gsl_matrix_get(p, i, j));
		}
		putchar('\n');
	}
}



	/***********************************************************************
	 * PROCEDURE: fill_bvec
	 *
	 * DESCRIPTION: Fill up the vector B, which is computed from 
	 *              observed magnitudes and weight.  This sits on the
	 *              right-hand side of the matrix equation.
	 *
	 *              There are two pieces to the B vector.  
	 *                 First, a set of ee values, 1 per image
	 *                 Next,  a set of ss values, 1 per star
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
fill_bvec
	(
	gsl_vector *b,         /* O: fill elements of this vector */
	int m,                 /* I: number of images */
	int n                  /* I: number of stars */
	)
{
	int i, j;
	double x;
	double temp1, temp2;

	/* 
	 * add up weights in each image; 
	 *   this includes an implicit sum over all filterchips
	 */
	for (i = 0; i < m; i++) {
		x = 0;
		for (j = 0; j < n; j++) {
			temp1 = gsl_matrix_get(weight, i, j);
			temp2 = gsl_matrix_get(mag, i, j);
			x += temp1*temp2;
		}
		gsl_vector_set(b, i, x);
	}

	/* 
	 * add up weights for each star; 
	 */
	for (j = 0; j < n; j++) {
		x = 0;
		for (i = 0; i < m; i++) {
			temp1 = gsl_matrix_get(weight, i, j);
			temp2 = gsl_matrix_get(mag, i, j);
			x += temp1*temp2;
		}
		gsl_vector_set(b, j + m, x);
	}

}


	/************************************************************************
	 * PROCEDURE: read_data
	 *
	 * DESCRIPTION: take the magnitude and uncertainty data, and figure
	 *              out how to turn it into weights for the
	 *              linear equations.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
read_data
	(
	struct s_star *star_arr,   /* I: array of stars */
	int m,                     /* I: number of images */
	int n                      /* I: number of stars */
	)
{
	int i, j;
	int num_measures;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;

	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			gsl_matrix_set(mag, i, j, 0.0);
			gsl_matrix_set(weight, i, j, 0.0);
		}
	}

	for (j = 0; j < n; j++) {
		star = &(star_arr[j]);	
		shAssert(star != NULL);
		mp = star->mp;
		num_measures = 0;
		while (mp != NULL) {
			measure = mp->measure;
			shAssert(measure != NULL);
			i = measure->image_id;
			gsl_matrix_set(mag, i, j, (double) measure->obs_mag);
			if ((star->include != 0) && (star->variable == 0) &&
			    (measure->include != 0) &&
				 (measure->image->include != 0)) {
				gsl_matrix_set(weight, i, j, (double) measure->w4);
			}
#ifdef DEBUG
			printf("  j %6d  i %6d  mag %6.3f  weight %6.3f \n", j, i, 
					gsl_matrix_get(mag, i, j), gsl_matrix_get(weight, i, j));
#endif

			num_measures++;
			mp = mp->next;
		}
		if (num_measures != star->num_measures) {
			fprintf(stderr, "read_data: num_measures %d but star_num %d ?!",
					num_measures, star->num_measures);
			exit(0);
		}
	}

}



	/***********************************************************************
	 * PROCEDURE: fill_matrix
	 *
	 * DESCRIPTION: Fill up the big matrix with properly weighted quantities.
	 *
	 *              The matrix can be broken into four "pieces"
	 *
	 *                   A      B      
	 *                   
	 *                   C      D 
	 *
	 *                   
	 *              where piece "A"  sum over all stars for each image
	 *                          "B"  weights of each image, star
	 *                          "C"  same as "B"
	 *                          "D"  sum over all images for each star
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
fill_matrix
	(
	gsl_matrix *p,        /* O: fill the elements of this matrix */
	int m,                /* I: number of images */
	int n                 /* I: number of stars */
	)
{
	int i, j;
	double x;

	for (i = 0; i < m + n; i++) {
		for (j = 0; j < m + n; j++) { 	
			gsl_matrix_set(p, i, j, 0.0);
		}
	}

	/* fill up the real matrix in four steps, one per piece */

	/*   piece "A": sum weights of all stars in each image */
	for (i = 0; i < m; i++) {
		x = 0;
		for (j = 0; j < n; j++) {
			x += gsl_matrix_get(weight, i, j);
		}
		gsl_matrix_set(p, i, i, x);
	}

	/*   pieces "B" and "C": weight for each individual star, image */
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			gsl_matrix_set(p, i, j + m, gsl_matrix_get(weight, i, j));
			gsl_matrix_set(p, j + m, i, gsl_matrix_get(weight, i, j));
		}
	}

	/*   piece "D": sum weights for all instances of each star */
	for (i = 0; i < n; i++) {
		x = 0;
		for (j = 0; j < m; j++) {
			x += gsl_matrix_get(weight, j, i);
		}
		gsl_matrix_set(p, i + m, i + m, x);
	}


#ifdef FORCE_STAR_MAG
	/* now we force the first star to have instrumental mag 0 */
	for (i = 0; i < m + n; i++) {
		gsl_matrix_set(p, i, m, 0.0);
	}
#endif

}


	/************************************************************************
	 * PROCEDURE: copy_matrix
	 *
	 * DESCRIPTION: copy the values of one matrix to another one
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
copy_matrix
	(
	gsl_matrix *p,         /* I: copy FROM this matrix ... */
	gsl_matrix *q,         /* O: INTO this matrix */
	int m,              /* I: number of rows in matrices */
	int n               /* I: number of columns in matrices */
	)
{
	int i, j;

	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			gsl_matrix_set(q, i, j, gsl_matrix_get(p, i, j));
		}
	}
}


	/***********************************************************************
	 * PROCEDURE: normalize
	 *
	 * DESCRIPTION: add a constant to all magnitudes 
	 *              (and to the x[] vector as well) so that the brightest 
	 *              star has the given mean magnitude value.
	 *              And also SUBTRACT this same constant to the image
	 *              adjustment factors so that we restore the 
	 *              equation connecting raw and corrected magnitudes
	 *              via the image adjustments.
	 *
	 * RETURNS:
	 *    nothing
	 *
	 */

static void
normalize
	(
	int m, 
	int n, 
	double value
	)
{
	int i, j;
	double min, diff, temp;

	min = 1.0e6;
	for (i = 0; i < n; i++) {
		if ((gsl_vector_get(mag0, i) < min) && 
		    (gsl_vector_get(mag0, i) != NODATA)) {
			min = gsl_vector_get(mag0, i);
		}
	}
	diff = (double) (value - min);
#ifdef DEBUG
	printf("normalize: diff is is %6.3f \n", diff);
#endif

	for (i = 0; i < n; i++) {
		if (gsl_vector_get(mag0, i) != NODATA) {
			temp = gsl_vector_get(mag0, i) + diff;
			gsl_vector_set(mag0, i, temp);
		}
		for (j = 0; j < m; j++) {
			if (gsl_matrix_get(Mes, j, i) != NODATA) {
				temp = gsl_matrix_get(Mes, j, i) + diff;
				gsl_matrix_set(Mes, j, i, temp);
			}
		}
	}

	/* 
	 * now, shift the image adjustment magnitudes by this same amount
	 * but in the opposite direction, to preserve the connection
	 * between input and output magnitudes
	 */
	for (i = 0; i < m; i++) { 
		if (gsl_vector_get(em, i) != NODATA) {
			temp = gsl_vector_get(em, i) - diff;
			gsl_vector_set(em, i, temp);
		}
	}
			
}



	/***********************************************************************
	 * PROCEDURE: find_means
	 *
	 * DESCRIPTION: find the mean magnitude of each star, either
	 *
	 *                 a. by simply looking in the 'x' vector, for 
	 *                    stars with non-zero weights
	 *                    (we solved for the mean mags of these stars)
	 *
	 *               or 
	 *
	 *                 b. by calculating it from the weighted average of
	 *                    (corrected) mag on each night 
	 *                    (these are "bad" or "variable" stars)
	 *
	 *               If there is no information for a star, 
	 *               use the NODATA value as its mean mag 
	 *
	 *               Once we know the value for a star, set its entry
	 *               in the "mag0" array to that value.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
find_means
	(
	struct s_star *star_arr,      /* I: array of stars */
	int m,                        /* I: number of images */
	int n                         /* I: number of stars */
	)
{
	int i, index;
	double top, bot;
	struct s_measurep *mp;
	struct s_measure *measure;

	for (i = 0; i < n; i++) {
		if ((star_arr[i].include != 0) && (star_arr[i].variable == 0)) {
			gsl_vector_set(mag0, i, gsl_vector_get(x, i + m));
		}
		else if (star_arr[i].variable != 0) {
			top = 0.0;
			bot = 0.0;
			mp = star_arr[i].mp;
			while (mp != NULL) {
				measure = mp->measure;
				shAssert(measure != NULL);

				index = measure->image_id;
				if (gsl_matrix_get(Mes, index, i) == NODATA) {
					mp = mp->next;
					continue;
				}
				top += gsl_matrix_get(Mes, index, i)*measure->w4;
				bot += measure->w4;
				mp = mp->next;
			}

			if (bot > 0.0) {
				gsl_vector_set(mag0, i, top/bot);
			}
			else {
				gsl_vector_set(mag0, i, NODATA);
			}

		}
	}
}


	/********************************************************************
	 * PROCEDURE: find_sigm
	 *
	 * DESCRIPTION: Return an estimate of the uncertainty sigma(m(e,s)) 
	 *              for star s in image e, which is only PART of the 
	 *              total error estimate sigma(M(e,s)).  
	 *              Find this value in two ways:
	 * 
	 *                1. try to find the average value of sigma(m0) for
	 *                          m0 ~= m(e,s)
	 *
	 *                2. if impossible, then use a value calculated
	 *                   from the weight of the value for m(e,s)
	 *                   since it is really based on input estimate
	 *                   of uncertainty, like so:
 	 *                      sig(m(e,s)) = 1.0/sqrt(weight(e,s))
	 *
	 *              If neither is possible, then the uncertainty
	 *              will be NODATA
	 *
	 * RETURNS:
	 *    the value of uncertainty in mean magnitude
	 */

static double
find_sigm
	(
	int n,              /* I: number of stars overall in solution */
	int e,              /* I: we are interested in a star in this image ... */
	int s,              /* I: ... and it's star 's' in this image */
	double mag_match,   /* I: how similar in mag must star be to act */
	                    /*      as stand-in to variable? */
	int this_id,        /* I: the ID of the star of interest */
	struct s_sort_mag *smag_array,   /* I: array of (mag0, sigm0) pairs */
	                                 /*    already sorted by mag0 */
	int num_sort        /* I: number of elemments in smag_array */

	)
{
	int num;
	int start_index, end_index, n_index;
	double dist, sum, wt, ret, this_mag;
   double temp, other_star_sigm0;


	/* first, try to find a mean m0 whose magnitude is close to 
	   m(e,s).  use a weighted average of the nearby sigm0. */
	sum = 0.0;
	wt = 0.0;
	num = 0;
	this_mag = gsl_matrix_get(Mes, e, s);
#ifdef DEBUG2
	printf("   find_sigm: star %3d in image %3d ID %5d mag %6.3f \n", 
					s, e, this_id, this_mag);
#endif

	if (find_neighbor_indices(smag_array, num_sort, this_mag, 
					mag_match, &start_index, &end_index) != 0) {
			fprintf(stderr, "find_sigm: find_neighbor_indices fails \n");
			exit(1);
	}
#ifdef DEBUG2
	printf("   find_neighbor_indices %5d %5d = %5d \n",
				start_index, end_index, (1 + end_index - start_index));
#endif
	for (n_index = start_index; n_index <= end_index; n_index++) {

		/* exclude the star itself from these calculations */
		if (smag_array[n_index].star_id == this_id) {
#ifdef DEBUG2
			printf("   skipping smag_array[%5d] has ID %5d \n",
						n_index, this_id);
#endif
			continue;
		}

		temp = smag_array[n_index].mag0 - this_mag;
		dist = fabs(temp);
#ifdef DEBUG2
		printf("       star %3d has mag0 %6.2f  target %3d has Mes %6.2f \n",
						n_index, smag_array[n_index].mag0, s, this_mag);
#endif
		/* 
		 * fixed bug here by adding extra check on sigm0 validity.
		 *   MWR 4/8/2006
		 */
		other_star_sigm0 = smag_array[n_index].sigm0;
#ifdef DEBUG2
		printf("  other_mag0 %7.3f other_sigm0 %7.3f \n", 
					smag_array[n_index].mag0, other_star_sigm0);
#endif
		if (other_star_sigm0 != NODATA) {
			if (dist < TINY) {
				dist = TINY;
			}
			sum += other_star_sigm0/dist;
#ifdef DEBUG2
			printf(" add to sum now num %6d %6.2f\n", (num+1), sum);
#endif
			wt += 1.0/dist;
			num++;
		}
#ifdef DEBUG2
		else {
			printf(" no good, ignore \n");
		}
#endif
	}
	if (num > 0) {
		ret = sum/wt;
#ifdef DEBUG2
		printf(" return sigm %6.3f  based on %5d stars \n", ret, num);
#endif
		return(ret);
	}

	/* second best: use the weight of instrumental mag(e,s) as an
	   estimator */
	if ((temp = gsl_matrix_get(weight, e, s)) > 0.0) {
		ret = 1.0/sqrt(temp);
		return(ret);
	}

	/* nothing left to do but return NODATA */
	ret = NODATA;
	return(ret);
}



	/********************************************************************
	 * PROCEDURE: build_sort_mag_array
	 *
	 * DESCRIPTION: Allocate space for an array which consists of 
	 *              simple structures.  Fill the structures with 
	 *              mag0 and sigm0 values from all the stars,
	 *              but exclude stars with 
	 *              values of NODATA.
	 *
	 *              Then sort the array by "mag0".
	 *              
	 *              We will use this array in the "find_sigm()" 
	 *              function to locate stars with magnitudes similar
	 *              to that of some given star.
	 *
	 * RETURNS:
	 *    pointer to the sorted array
	 */

static struct s_sort_mag *
build_sort_mag_array
	(
	struct s_star *star_arr,   /* I: array of all stars */
	int num_stars,      /* I: number of stars in solution */
	int *num_in_array   /* O: the number of stars placed into array */
	)
{
	int i, num_sort;
	struct s_sort_mag *sort_mag_array;

	/*
 	 * as part of job of identifying stars with similar magnitudes,
 	 * we will build an array and sort it by magnitude.  
 	 */
	if ((sort_mag_array = (struct s_sort_mag *) calloc(num_stars, sizeof(struct s_sort_mag))) == NULL) {
		fprintf(stderr, "build_sort_mag_array: can't calloc for %d s_sort_mag \n", num_stars);
		exit(1);
	}
	num_sort = 0;
	for (i = 0; i < num_stars; i++) {
		int this_id;
		double this_mag0, this_sigm0;

		this_id = star_arr[i].id;
		this_mag0 = gsl_vector_get(mag0, i);
		this_sigm0 = gsl_vector_get(sigm0, i);
		if ((this_mag0 != NODATA) && (this_sigm0 != NODATA)) {
			sort_mag_array[num_sort].star_id = this_id;
			sort_mag_array[num_sort].mag0 = this_mag0;
			sort_mag_array[num_sort].sigm0 = this_sigm0;
#ifdef DEBUG
			printf("   sort_mag_array[%5d] has mag0 %9.5f sigm0 %9.5f \n", 
						num_sort, sort_mag_array[num_sort].mag0, 
						sort_mag_array[num_sort].sigm0);
#endif
			num_sort++;
		}
	}
	qsort((char *) sort_mag_array, num_sort, sizeof(struct s_sort_mag), (PFI) comp_sortmag);
#ifdef DEBUG2
	for (i = 0; i < num_sort; i++) {
		printf("   sorted mags_array[%5d]  %5d %6.3f %6.3f \n", 
				i, sort_mag_array[i].star_id, 
				sort_mag_array[i].mag0, sort_mag_array[i].sigm0);
	}
#endif

	*num_in_array = num_sort;
	return(sort_mag_array);
}


	/********************************************************************
	 * PROCEDURE: find_neighbor_indices
	 *
	 * DESCRIPTION: Given an array of s_sort_mag structures which
	 *              are sorted by "mag0", a particular value
	 *              of magnitude, and a range in magnitude,
	 *              use a binary search to find
	 *              the indices in the array of stars which are
	 *              "neighbors" of the given magnitude.
	 *
	 *                 start_index   is index of element in array
	 *                                 with (mag0 - mag_match)
	 *
	 *                 end_index     is index of element in array
	 *                                 with (mag0 + mag_match)
	 *
	 *              Place these two values into the final two
	 *              arguments of the function.
	 *
	 * RETURNS:
	 *    0                   if all goes well
	 *    1                   if an error occurs
	 */


static int
find_neighbor_indices
	(
	struct s_sort_mag *smag_array,  /* I: array already sorted by mag */
	int num_stars,                  /* I: number of stars in array */
	double this_mag,                /* I: look for mags near this value */
	double mag_range,               /* I: acceptable range from desired value */
	int *start_index,               /* O: index of brightest neighbor */
	int *end_index                  /* O: index of faintest neighbor */
	)
{
	int top, bottom, mid;
	double seek_mag;

	/* first, we search for the "brightest" neighbor */
	seek_mag = this_mag - mag_range;
#ifdef DEBUG2
	printf("find_neighbor_indices: looking for mag %6.3f \n", seek_mag);
#endif

	top = 0;
	bottom = num_stars - 1;

	while (bottom - top >= 2) {
		mid = (top + bottom) / 2;
#ifdef DEBUG2
		printf("  ar[%5d] %6.3f  ar[%5d] %6.3f  ar[%5d] %6.3f \n",
				top, smag_array[top].mag0,
				mid, smag_array[mid].mag0,
				bottom, smag_array[bottom].mag0);
#endif
		if (smag_array[mid].mag0 < seek_mag) {
			top = mid;
		}
		else {
			bottom = mid;
		}
	}
#ifdef DEBUG2
	printf("  ar[%5d] %6.3f                 ar[%5d] %6.3f \n",
				top, smag_array[top].mag0,
				bottom, smag_array[bottom].mag0);
#endif

	/* 
 	 * when we reach this point, the item we seek is either "top"
 	 *    or "bottom" in the array
 	 */
	if (smag_array[top].mag0 < seek_mag) {
#ifdef DEBUG2
		printf("     bright is bottom ar[%5d] %6.3f \n",
				bottom, smag_array[bottom].mag0);
#endif
		*start_index = bottom;
	}
	else {
#ifdef DEBUG2
		printf("     bright is top    ar[%5d] %6.3f \n",
				top, smag_array[top].mag0);
#endif
		*start_index = top;
	}

	/* now, repeat the process to find the "faintest neighbor" */
	seek_mag = this_mag + mag_range;

	top = 0;
	bottom = num_stars - 1;

	while (bottom - top >= 2) {
		mid = (top + bottom) / 2;
#ifdef DEBUG2
		printf("  ar[%5d] %6.3f  ar[%5d] %6.3f  ar[%5d] %6.3f \n",
				top, smag_array[top].mag0,
				mid, smag_array[mid].mag0,
				bottom, smag_array[bottom].mag0);
#endif
		if (smag_array[mid].mag0 < seek_mag) {
			top = mid;
		}
		else {
			bottom = mid;
		}
	}
#ifdef DEBUG2
	printf("  ar[%5d] %6.3f                 ar[%5d] %6.3f \n",
				top, smag_array[top].mag0,
				bottom, smag_array[bottom].mag0);
#endif

	/* 
 	 * when we reach this point, the item we seek is either "top"
 	 *    or "bottom" in the array
 	 */
	if (smag_array[top].mag0 < seek_mag) {
#ifdef DEBUG2
		printf("     faint is  bottom ar[%5d] %6.3f \n",
				bottom, smag_array[bottom].mag0);
#endif
		*end_index = bottom;
	}
	else {
#ifdef DEBUG2
		printf("     faint is  top    ar[%5d] %6.3f \n",
				top, smag_array[top].mag0);
#endif
		*end_index = top;
	}

	return(0);
}

	/***********************************************************************
	 * PROCEDURE: write_images
	 *
	 * DESCRIPTION: Write output about images ONLY to the given file 
	 *              (or stdout, if the given string is "").
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_images
	(
	struct s_image *image_arr,       /* I: array of images to print */
	int num_images,                  /* I: number of images in array */
	char *file                       /* I: output file name */
	)
{
	int i;
	FILE *fp;
	struct s_image *im;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}

	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include == 0) {
			continue;
		}
		im = &(image_arr[i]);
		fprintf(fp, "%-4d %30s %6.1f  %14.7f ",
			im->index, im->name, im->exptime, im->jd);
		fprintf(fp, "%5.3f %5.3f %5.3f\n", 
							gsl_vector_get(em, i), 
							gsl_vector_get(sigem, i), 
							gsl_vector_get(sigemn, i));
	}
	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}


	/***********************************************************************
	 * PROCEDURE: write_stars
	 *
	 * DESCRIPTION: Write information about stars ONLY, one line per
	 *              star per image, to the given file 
	 *              (or stdout if file == "").  
    *
	 *              We add the given "delta_mag" to all magnitude values
	 *              before printing them out.  It will be zero
	 *              unless the user has given the "mrange" commend-line
	 *              option.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_stars
	(
	struct s_star *star_arr,   /* I: array of stars to print */
	int num_stars,             /* I: number of stars in array */
	char *file,                /* I: output file name or "" for stdout */
	double delta_mag           /* I: add this to all output mag values */
	)
{
	int i, index;
	FILE *fp;
	struct s_star *star;
	struct s_image *im;
	struct s_measurep *mp;
	struct s_measure *measure;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(0);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(0);
	}

	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include == 0) {
			continue;
		}
		star = &(star_arr[i]);
		mp = star->mp;
		while (mp != NULL) {
			measure = mp->measure;
			shAssert(mp != NULL);

			im = measure->image;
			shAssert(im != NULL);
			if (im->include != 0) {
				index = im->index;
				fprintf(fp, "%-6d %9.5f %9.5f %4d %11.5f ",
						star->id, star->ra, star->dec, index, im->jd);
				fprintf(fp, " %6.3f %6.3f %5.3f ", 
						gsl_vector_get(mag0, i) + delta_mag,
						gsl_matrix_get(Mes, index, i) + delta_mag, 
						gsl_matrix_get(sigMes, index, i));
				fprintf(fp, " %2d %7.2f \n",
						star->variable, star->nsigma);
			}
			mp = mp->next;
		}
	}
	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}





	/************************************************************************
	 * PROCEDURE: write_sigma
	 *
	 * DESCRIPTION: Write information about sigma0 vs. mag0 
	 *              (standard deviation from mean mag, versus mean mag)
	 *              to the given file (or stdout if the given file == "") 
	 *
	 *              We add the given "delta_mag" to all magnitude values
	 *              before printing them out.  It will be zero
	 *              unless the user has given the "mrange" commend-line
	 *              option.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_sigma
	(
	struct s_star *star_arr, /* I: array of stars to print */
	int num_stars,           /* I: number of stars in array */
	char *file,              /* I: output file name, or "" for stdout */
	double delta_mag         /* I: add this to all output magnitudes */
	)
{
	int i;
	FILE *fp;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}

	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include != 0)  {
			fprintf(fp, "%6d %6.3f %6.3f %2d %7.2f %12.5f %12.5f\n", 
						star_arr[i].id, 
						star_arr[i].mag0 + delta_mag,
						star_arr[i].sigma,
						star_arr[i].variable,
						star_arr[i].nsigma,
						star_arr[i].ra,
						star_arr[i].dec);
		}
	}

	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}




#if 0
	/************************************************************************
	 * PROCEDURE: calc_model_mag
	 *
	 * DESCRIPTION: We have finished finding the photometric solution,
	 *              so we have the image adjustments, mean star magnitudes,
	 *              and chip offsets.  Here, we figure out the magnitude
	 *              a star _ought_ to have, based on its image, star,
	 *              chip, etc.  We go back to the x[] vector values
	 *              for all these calculations.
	 *
	 *              The value returned can be compared to the actual
	 *              measured magnitude; in a perfect world, they
	 *              would be the same.  In the real world, we expect
	 *              small residuals which (if we have done a good 
	 *              job in the model) will be randomly scattered
	 *              around zero.
	 *
	 * RETURNS:
	 *     magnitude of star from solution
	 *     NODATA         if this isn't a valid measurement
	 */

static double
calc_model_mag
	(
	int num_images,        /* I: total number of images in solution */
	int num_stars,         /* I: total number of stars in solution */
	int image_index,       /* I: index of image, 1 ... num_images */
	int star_index         /* I: index of star, 1 .. num_stars */
	)
{
	double star_mag0, mag_correction;
	double model_mag;

	/* we don't (yet) consider whether this is a valid measurement */

	star_mag0 = x[num_images + star_index];
	mag_correction = calc_model_correction(num_images, num_stars,
			              image_index, star_index);

	if (mag_correction == NODATA) {
		model_mag = NODATA;
	}
	else {
		model_mag = star_mag0 + mag_correction;
	}

	return(model_mag);
}

#endif



	/************************************************************************
	 * PROCEDURE: calc_model_correction
	 
	 * DESCRIPTION: We have finished finding the photometric solution,
	 *              so we have the image adjustments, mean star magnitudes,
	 *              and chip offsets.  Here, we figure out the total
	 *              size of all the adjustments, due to image zeropoint,
	 *              chip offset, filter mismatch, etc.
	 *
	 *              The result should be used in this sense:
	 *
	 *                  obs_mag  =   true_mag  +  corrections
	 *
	 *
	 * RETURNS:
	 *     corrections to star's mag
	 *     NODATA       if we can't calculate it
	 */

static double
calc_model_correction
	(
	int num_images,        /* I: total number of images in solution */
	int num_stars,         /* I: total number of stars in solution */
	int image_index,       /* I: index of image, 1 ... num_images */
	int star_index         /* I: index of star, 1 .. num_stars */
	)
{
	double image_em;
	double mag_correction;

	/* we don't (yet) consider whether this is a valid measurement */
	image_em = gsl_vector_get(x, image_index);

	mag_correction = image_em;

	return(mag_correction);
}



	/************************************************************************
	 * PROCEDURE: copy_image_array
	 
	 * DESCRIPTION: Given a pointer to array of images, 
	 *              create a new array of s_image structures,
	 *              which has exactly the same properties as the
	 *              original.
	 *
	 *              This routine allocates the space for the new array,
	 *              and for all items within it.
	 *
	 * RETURNS:
	 *     pointer to new array         if all goes well
	 *     NULL                         if allocation fails
	 */

static struct s_image *
copy_image_array
	(
	struct s_image *orig_array,  /* I: pointer to original array we will copy */
	int num_orig_images    /* I: number of elements in original array */
	)
{
	int i;
	struct s_image *new_array;

	shAssert(orig_array != NULL);

	/* 
    * first, allocate the space necessary for the s_image structures 
    * themselves.
    */
	if ((new_array = (struct s_image *) calloc(num_orig_images, sizeof(struct s_image))) == NULL) {
		fprintf(stderr, "copy_image_array: can't calloc for %d images \n", 
					num_orig_images);
		return(NULL);
	}

	/* 
 	 * walk through array and copy values for each element, 
 	 *    including a new linked list of s_measurep structures
 	 */
	for (i = 0; i < num_orig_images; i++) {
		int i_measure;
		struct s_image *orig, *new;
		struct s_measurep *orig_mp, *prev_mp;

		orig = &(orig_array[i]);
		new = &(new_array[i]);

		new->index = orig->index;
		strncpy(new->name, orig->name, STRLEN - 1); 
		new->name[STRLEN - 1] = '\0';
		strncpy(new->object, orig->object, STRLEN - 1); 
		new->object[STRLEN - 1] = '\0';
		new->exptime = orig->exptime;
		new->jd = orig->jd;
		new->num_measures = orig->num_measures;
		orig_mp = orig->mp;

		/* make a new list of pointers to the original's measures */
		prev_mp = NULL;
		for (i_measure = 0; i_measure < new->num_measures; i_measure++) {
			struct s_measurep *mpp;

			if ((mpp = (struct s_measurep *) malloc(sizeof(struct s_measurep))) == NULL) {
				fprintf(stderr, "copy_images: can't malloc for measure_p %d \n", i_measure);
				return(NULL);
			}
			shAssert(orig_mp != NULL);
			mpp->measure = orig_mp->measure;
			mpp->next = NULL;
			if (i_measure == 0) {
				new->mp = mpp;
			}
			else {
				prev_mp->next = mpp;
			}
			prev_mp = mpp;

			orig_mp = orig_mp->next;
		}

		new->num_match = orig->num_match;
		new->xshift = orig->xshift;
		new->yshift = orig->yshift;
		new->include = orig->include;

	}

	return(new_array);
}




#ifdef DEBUG
	/************************************************************************
	 * PROCEDURE: print_image_array
	 
	 * DESCRIPTION: Given a pointer to array of images, 
	 *              print out one line of basic information per image,
	 *              and print out terse info on each measure in each image.
	 *
	 * RETURNS:
	 * 	     nothing
	 */

static void
print_image_array
	(
	struct s_image *im_array, /* I: pointer to array of images */
	int num_images            /* I: number of elements in array */
	)
{
	int i;
	struct s_image *ip;
	struct s_measure *m;
	struct s_measurep *mp;

	for (i = 0; i < num_images; i++) {
		ip = &(im_array[i]);
		shAssert(ip != NULL);

		printf(" %5d  %7.2f %15.6f  %5d  %5d %6.2f %6.2f  %d \n",
			ip->index, ip->exptime, ip->jd, ip->num_measures,
			ip->num_match, ip->xshift, ip->yshift, ip->include);

		/* now print out one terse line per measurep in the image */
		for (mp = ip->mp; mp != NULL; mp = mp->next) {
			m = mp->measure;
			printf("   %8d %5d %5d  %6.3f %6.3f %d %9.4e  %d \n",
						m->measure_id, m->image_id, m->star_id,
						m->obs_mag, m->obs_sigma, m->w3, m->w4, m->include);
		}
	}

}
#endif


	/************************************************************************
	 * PROCEDURE: copy_star_subset
	 
	 * DESCRIPTION: Given a pointer to array of stars, 
	 *              create a new array of s_star structures,
	 *              including only those which meet a criterion
	 *              set by the user:
	 *                  only the brightest n_bright stars
	 *              
	 *              Each copied star should have
	 *              exactly the same properties as the original.
	 *
	 *              This routine allocates the space for the new array,
	 *              and for all items within it.
	 *
	 * RETURNS:
	 *     pointer to new array         if all goes well
	 *     NULL                         if allocation fails
	 */

static struct s_star *
copy_star_subset
	(
	struct s_star *orig_array,  /* I: pointer to original array we will copy */
	int num_orig_stars,    /* I: number of elements in original array */
	int n_bright           /* I: copy the n_bright brightest */
	)
{
	int i, new_count;
	double cutoff_mag;
	struct s_star *new_array, *new_sp;

	shAssert(orig_array != NULL);
	shAssert((n_bright > 0) && (n_bright < num_orig_stars));

	/* 
 	 * first, we need to determine the magnitude cutoff: the magnitude
 	 *    of the n_bright'th brightest star.  So, our first job is
 	 *    to compute this cutoff magnitude.
 	 */
	if ((cutoff_mag = calc_cutoff_mag(orig_array, num_orig_stars, n_bright)) == NODATA) {
		fprintf(stderr, "copy_star_subset: can't compute cutoff_mag ?! \n");
		return(NULL);
	}
#ifdef DEBUG
	printf(" copy_star_subset: cutoff_mag is %6.3f \n", cutoff_mag);
#endif

	/* 
    * allocate the space necessary for a copy of only the
    * n_bright brightest stars
    *
    */
	if ((new_array = (struct s_star *) calloc(n_bright, sizeof(struct s_star))) == NULL) {
		fprintf(stderr, "copy_star_array: can't calloc for %d stars \n", 
					num_orig_stars);
		return(NULL);
	}

	/* 
 	 * Now, we walk through the original array, comparing the 'mag0'
 	 *    value of each star to the 'cutoff_mag'.  Stars which
 	 *    are brighter than the cutoff_mag will be copied into
 	 *    the new array.
 	 *
 	 * However, it's possible that a number of stars might have
 	 *    exactly the same value of 'mag0' = 'cutoff_mag', 
 	 *    leading to too many stars being copied.  So, once we reach 
 	 *    the desired number 'n_bright' stars, stop copying.
 	 */
	new_count = 0;
	new_sp = &(new_array[new_count]);
	for (i = 0; i < num_orig_stars; i++) {
		struct s_star *orig_sp;

		orig_sp = &(orig_array[i]);
		if (orig_sp->mag0 > cutoff_mag) {
			continue;
		}

		if (copy_one_star(orig_sp, new_sp) != 0) {
			fprintf(stderr, "copy_star_subset: copy_one_star fails for star %d \n", i);
			return(NULL);
		}

		new_count++;
		new_sp = &(new_array[new_count]);

		if (new_count >= n_bright) {
			break;
		}
	}


	return(new_array);
}


	/************************************************************************
	 * PROCEDURE: copy_one_star
	 
	 * DESCRIPTION: Given a pointer to two existing s_star structures,
	 *              copy all the information from the first to the second;
	 *              that means walking through the list of s_measurep
	 *              structures, allocating memory for a new linked list,
	 *              and copying pointers to each measurement.
	 *
	 * 	
	 *
	 * RETURNS:
	 *     0                          if all goes well
	 *     1                          if an error occurs
	 */

static int
copy_one_star
	(
	struct s_star *from,   /* I: copy from this structure ... */
	struct s_star *to      /* I: ... into this structure */
	)
{
	int i;
	struct s_measurep *from_mp, *prev_mp, *new_mp;

	shAssert(from != NULL);
	shAssert(to != NULL);

	to->id = from->id;
	to->ra = from->ra;
	to->dec = from->dec;
	to->mag0 = from->mag0;
	to->sigma = from->sigma;
	to->nsigma = from->nsigma;
	to->num_measures = from->num_measures;

	from_mp = from->mp;
	prev_mp = NULL;
	for (i = 0; i < from->num_measures; i++) {
		if ((new_mp = (struct s_measurep *) malloc(sizeof(struct s_measurep))) == NULL) {
			fprintf(stderr, "copy_one_star: malloc fails for measurep %d \n", i);
			return(1);
		}
		new_mp->measure = from_mp->measure;
		new_mp->next = NULL;
		if (i == 0) {
			to->mp = new_mp;
		}
		else {
			prev_mp->next = new_mp;
		}

		prev_mp = new_mp;
		from_mp = from_mp->next;
	}

	to->include = from->include;
	to->variable = from->variable;

#ifdef DEBUG
	{
	int i;
	struct s_measurep *from_mp, *to_mp;

	printf(" copy_one_star: id           %5d %5d \n", from->id, to->id);
	printf(" copy_one_star: ra           %12.5f %12.5f \n", from->ra, to->ra);
	printf(" copy_one_star: dec          %12.5f %12.5f \n", from->dec, to->dec);
	printf(" copy_one_star: mag0         %6.3f %6.3f \n", from->mag0, to->mag0);
	printf(" copy_one_star: sigma        %6.3f %6.3f \n", from->sigma, to->sigma);
	printf(" copy_one_star: nsigma       %5.2f %5.2f \n", from->nsigma, to->nsigma);
	printf(" copy_one_star: num_measures %5d %5d \n", from->num_measures, to->num_measures);

	from_mp = from->mp;
	to_mp = to->mp;
	for (i = 0; i < from->num_measures; i++) {
		printf("                mp %5d    %10d %10d \n", 
				i, from_mp->measure->measure_id, to_mp->measure->measure_id);
		from_mp = from_mp->next;
		to_mp = to_mp->next;
	}

	printf(" copy_one_star: include      %5d %5d \n", from->include, to->include);
	printf(" copy_one_star: variable     %5d %5d \n", from->variable, to->variable);
	}
#endif

	return(0);
}


   /**************************************************************
    * PROCEDURE: comp_double
    *
    * DESCRIPTION: Compare the values of two double-precision numbers
    *
    * RETURNS:
    *       -1 if the first number has a value smaller than the second
    *        0                                 equal to 
    *       +1                                 larger than
    */

static int
comp_double
   (
   double *d1,
   double *d2
	)
{
   if ((*d1) < (*d2)) {
      return(-1);
   }
   if ((*d1) > (*d2)) {
      return(1);
   }
   return(0);
}


   /**************************************************************
    * PROCEDURE: comp_sortmag
    *
    * DESCRIPTION: Compare two s_sort_mag structures by their
    *              magnitude fields.  
    *
    * RETURNS:
    *       -1 if the first struct has a mag0  smaller than the second
    *        0                                 equal to 
    *       +1                                 larger than
    */

static int
comp_sortmag
   (
   struct s_sort_mag *s1,
   struct s_sort_mag *s2
	)
{
   if ((s1->mag0) < (s2->mag0)) {
      return(-1);
   }
   if ((s1->mag0) > (s2->mag0)) {
      return(1);
   }
   return(0);
}




	/************************************************************************
	 * PROCEDURE: calc_cutoff_mag
	 
	 * DESCRIPTION: Given a pointer to array of stars, 
	 *              the number of stars, and the number 'n_bright', 
	 *              of stars desired in a new bright subset, 
	 *              determine the average magnitude of the 
	 *              n_bright'th star. 
	 *
	 *              We can use that to select stars for our 
	 *              new, bright-only subset.
	 *
	 * RETURNS:
	 *     magnitude of n_bright'est star    if all goes well
	 *     NODATA                            if error occurs
	 */

static double
calc_cutoff_mag
	(
	struct s_star *star_array,  /* I: pointer to array of stars */
	int num_stars,              /* I: number of elements in array */
	int n_bright                /* I: rank of desired star by brightness */
	)
{
	int i, j;
	double cutoff_mag;
	double *mag_array;

	shAssert(star_array != NULL);
	shAssert((n_bright > 0) && (n_bright < num_stars));

	/* 
 	 * Make a quick scan through the entire
 	 *    array, computing an "average" input magnitude.  Ignore
 	 *    any measurements with "include == 0", or with NODATA.
 	 *    Set the value of a temporary "mag_array[]" element to this
 	 *    average value; we will sort that array in a moment.
 	 *
 	 * Also, set the 'mag0' field of each element of the input star_array
 	 *    to the average magnitude for its measurements.
 	 *    This will help us to pick out the subset for calculations
 	 *    in the near future.
 	 */

	if ((mag_array = (double *) calloc(num_stars, sizeof(double))) == NULL) {
		fprintf(stderr, "calc_cutoff_mag: can't calloc for %d mag_array \n", 
					num_stars);
		return(NODATA);
	}

	for (i = 0; i < num_stars; i++) {
		double num, sum, avg_mag;
		struct s_star *orig_sp;
		struct s_measure *m;
		struct s_measurep *mp;

		orig_sp = &(star_array[i]);

		num = 0.0;
		sum = 0.0;
		mp = orig_sp->mp;
		for (j = 0; j < orig_sp->num_measures; j++) {

			shAssert(mp != NULL);
			m = mp->measure;
			shAssert(m != NULL);

			if (m->obs_mag != NODATA) {
				num++;
				sum += m->obs_mag;
			}
			mp = mp->next;
		}
		if (num > 0) {
			avg_mag = sum / num;
		}
		else {
			avg_mag = NODATA;
		}
		mag_array[i] = avg_mag;
		orig_sp->mag0 = avg_mag;
#ifdef DEBUG
		printf(" calc_cutoff_mag: star %5d has num %5.0f  avg %6.3lf \n",
					orig_sp->id, num, avg_mag);
#endif
	}

	/* now sort the array of magnitudes */
	qsort((char *) mag_array, num_stars, sizeof(double),
				(PFI) comp_double);
#ifdef DEBUG
	for (i = 0; i < num_stars; i++) {
		printf(" calc_cutoff_mag: sorted star %5d has mag %6.3f \n",
				i, mag_array[i]);
	}
#endif

	/* determine the magnitude of the n_bright'th brighest star */
	cutoff_mag = mag_array[n_bright - 1];
#ifdef DEBUG
	printf(" calc_cutoff_mag: n_bright %5d  star has mag %6.3f \n",
			n_bright, cutoff_mag);
#endif
	
	return(cutoff_mag);
}





	/************************************************************************
	 * PROCEDURE: weed_image_array
	 
	 * DESCRIPTION: Given a pointer to an array of images and number
	 *              of images, and a small array of stars, number of stars
	 *              in that small array, and the number of stars in the
	 *              full-size original list of all stars,
	 *              we walk through the images, getting rid of
	 *              measurements of stars which are NOT in the given
	 *              array.   
	 *
	 *              The goal is to create self-consistent arrays of
	 *              images and stars: only these stars are listed in
	 *              all the images.
	 *
	 * RETURNS:
	 *     0                         if all goes well
	 *     1                         if an error occurs     
	 */

static int
weed_image_array
	(
	struct s_image *image_array,     /* I: pointer to array of images */
	int num_images,                  /* I: number of images */
	struct s_star *star_array,       /* I: pointer to small array of stars */
	int num_stars,                   /* I: number of stars in small array */
	int num_original_stars           /* I: number of stars in original, full array */
	)
{
	int i, i_image, *hash;
	int this_star;
	

	shAssert(image_array != NULL);
	shAssert(star_array != NULL);

	/* 
 	 * create a little hash table so that we can quickly tell if a given
 	 *    star_id is in the given star_array or not.  Initialize
 	 *    all elements of the hash table to zero.
 	 */
	if ((hash = (int *) calloc(num_original_stars, sizeof(int))) == NULL) {
		fprintf(stderr, "weed_image_array: can't calloc for %d stars \n", 
					num_stars);
		return(1);
	}
	for (i = 0; i < num_original_stars; i++) {
		hash[i] = 0;
	}

	/* 
 	 * now set elements of hash table corresponding to stars in our
 	 *     small subset to 1.  
 	 */
	for (i = 0; i < num_stars; i++) {
		hash[star_array[i].id] = 1;
	}
#ifdef DEBUG2
	{
		int ii;
		for (ii = 0; ii < num_original_stars; ii++) {
			printf("     hash[ %5d ] = %d \n", ii, hash[ii]);
		}
	}
#endif

	for (i_image = 0; i_image < num_images; i_image++) {
		int new_num_star;
#ifdef DEBUG2
		int mp_cur_count, mp_start_count;
#endif
		struct s_image *ip;
		struct s_measurep *mp, *prev_mp;

		ip = &(image_array[i_image]);
#ifdef DEBUG2
		printf("weed_image_array: top of loop for image %d \n", i_image);
		mp_start_count = ip->num_measures;
		mp_cur_count = 0;
#endif
		new_num_star = 0;

		prev_mp = ip->mp;
		mp = ip->mp;

		while (mp != NULL) {
#ifdef DEBUG2
			printf("    top of loop, start_count %5d  cur_count %5d \n",
						mp_start_count, mp_cur_count);
#endif

			this_star = mp->measure->star_id;
#ifdef DEBUG2
			printf("   image %5d  mp %10d star %5d \n", 
					ip->index, mp->measure->measure_id, this_star);
#endif
			if (hash[this_star] == 1) {
				/* we want to keep this measurement, so don't delete it */
				prev_mp = mp;
				mp = mp->next;
				new_num_star++;
#ifdef DEBUG2
				printf("        hash found, keep it \n");
				fflush(stdout);
#endif
			}
			else {
				/* we want to remove this measurement */
				if (mp == ip->mp) {
					ip->mp = mp->next;
					free(mp);
					mp = ip->mp;
#ifdef DEBUG2
				printf("        not found, reset ip->mp to mp->next  \n");
				fflush(stdout);
#endif
				}
				else {
					prev_mp->next = mp->next;
					free(mp);
					mp = prev_mp->next;
#ifdef DEBUG2
				printf("        not found, set prev->next to mp->next  \n");
				fflush(stdout);
#endif
				}
			}

#ifdef DEBUG2
			mp_cur_count++;
#endif
		}

#ifdef DEBUG2
		printf("       end for image %5d has %5d stars now \n",
					i_image, new_num_star);
#endif

		/* re-set the value of num_measures in this image */
		ip->num_measures = new_num_star;
	}

	/* 
 	 * sanity check: walk through image array and verify that
 	 * 	each image has <= n_bright measurements
 	 */
	for (i_image = 0; i_image < num_images; i_image++) {
		int i_measure;
		struct s_image *ip;
		struct s_measurep *mp;

		ip = &(image_array[i_image]);
#ifdef DEBUG2
		printf("weed_image_array: top of loop for image %d \n", i_image);
#endif

		mp = ip->mp;
		for (i_measure = 0; i_measure < ip->num_measures; i++) {
			shAssert(mp != NULL);
#ifdef DEBUG2
			printf("    image %5d  measure %5d is star %5d \n",
					i_image, i_measure, mp->measure->star_id);
#endif
			mp = mp->next;
			if (mp == NULL) {
				break;
			}
		}

	}
			

	/* now free the space for the hash table */
	free(hash);

	return(0);
}


	/************************************************************************
	 * PROCEDURE: transfer_solution
	 
	 * DESCRIPTION: We have created a subset of all the stars and 
	 *              run the ensemble code on the subset.  Now,
	 *              we need to transfer the information from that
	 *              small solution into the arrays used for 
	 *              the full dataset.
	 *
	 * RETURNS:
	 *     0                         if all goes well
	 *     1                         if an error occurs     
	 */

static int
transfer_solution
	(
	struct s_image *full_image_array,  /* I: array of all images */
	int full_num_images,             /* I: number of all images */
	struct s_star *full_star_array,  /* I: array of all stars */
	int full_num_stars,              /* I: number of all stars */
	struct s_image *q_image_array,   /* I: array of images used in subset */
                                    /*      (same number as full_num_images) */
	struct s_star *q_star_array,     /* I: array of stars in subset */
	int q_num_stars                  /* I: number of stars in subset */
	)
{
	int i;
	gsl_vector *full_x;

	shAssert(full_image_array != NULL);
	shAssert(full_star_array != NULL);
	shAssert(q_image_array != NULL);
	shAssert(q_star_array != NULL);

	/* 
 	 * create a new GSL vector to hold the solutions for all images
 	 *    and all stars, not just the stars in the small subset.
 	 *    We will later cause the variable "x" to point to this
 	 *    new, full-size vector.
 	 */
	if ((full_x = gsl_vector_calloc(full_num_images + full_num_stars)) == NULL) {
		fprintf(stderr, "transfer_solution: can't allocate for vector full_x \n");
		exit(-1);
	}
	/* transfer image zeropoints from quick solution */
	for (i = 0; i < full_num_images; i++) {
		double em;
		em = gsl_vector_get(x, i);
		gsl_vector_set(full_x, i, em);
	}
	/* transfer star magnitudes from quick solution */
	for (i = 0; i < q_num_stars; i++) {
		int small_index, full_index;
		double val;

		small_index = full_num_images + i;
		val = gsl_vector_get(x, small_index);
		full_index = full_num_images + q_star_array[i].id;
		gsl_vector_set(full_x, full_index, val);
	}
	/* now delete the old GSL vector "x" ... */
	gsl_vector_free(x);
	/* ... and re-assign the variable "x" to the new full-sized vector */
	x = full_x;



	/*
 	 * delete the existing, subset-sized GSL matrices 
 	 *                  "mag" and "weight"
 	 * and re-allocate new, full-sized versions.  Then re-fill
 	 * those full-sized versions from the original, full-size input data
 	 */
	gsl_matrix_free(mag);
	gsl_matrix_free(weight);
	if ((mag = gsl_matrix_calloc(full_num_images, full_num_stars)) == NULL) {
		fprintf(stderr, "transfer_solution: can't allocate for matrix mag\n");
		exit(-1);
	}
	if ((weight = gsl_matrix_calloc(full_num_images, full_num_stars)) == NULL) {
		fprintf(stderr, "transfer_solution: can't allocate for matrix weight\n");
		exit(-1);
	}
	read_data(full_star_array, full_num_images, full_num_stars);

	/* 
 	 * go through the full-size star array and set the "variable" fields
 	 * 	as follows:
 	 *
 	 * 	  - if star was included in "quick" subset, variable = 0
 	 * 	  - if star was NOT      in "quick" setset, variable = 2
 	 */
	for (i = 0; i < full_num_stars; i++) {
		full_star_array[i].variable = 2;
	}
	for (i = 0; i < q_num_stars; i++) {
		int index;

		index = q_star_array[i].id;
		full_star_array[index].variable = 0;
	}


	return(0);
}

